/*
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Copyright (c) 2011 digitalSTROM.org, Zurich, Switzerland
 *  Author: Roman Khler <roman.koehler@aizo.com>
 *  Complete rewrite by: Andreas Brauchli <andreas.brauchli@aizo.com>
 */

var LOGFILE_NAME = 'timed-events.log',
    LOG = new Logger(LOGFILE_NAME),
    WEEK = ['MO','TU','WE','TH','FR','SA','SU'],
    daytimeEvents = {
        sunrise: 'sunrise',
        sunset: 'sunset'
    };

function pad(c) {
    var i = parseInt(c, 10);
    return (i < 10 ? '0' + i : c.toString());
}

function getDayIdx(d) {
    var i;
    for (i = 0; i < 7; ++i) {
        if (d == WEEK[i]) {
            return i;
        }
    }
    return -1;
}

/**
 * Returns an unused unique identifier (id)
 * @return {String} A unique string
 */
function getId() {
    var ndEvents = Property.getNode('events'),
        events,
        max,
        tmp,
        i;

    if (ndEvents) {
        events = ndEvents.getChildren();
        max = 0;
        while (ndEvents.getChild(max.toString())) {
            ++max;
        }
        return max.toString();

    } else {
        return "0";
    }
}

/**
 * Function that splits the daytime from the time string
 * @param time {String} The time string as saved
 * @return {Array|boolean|String}
 *          An array of length 3 containing:
 *              {String} the daytime event,
 *              {boolean} sign of the delay
 *              {String} delay
 *          The time string as input if time is not a daytime
 *          false if the time input is not a valid daytime
 */
function getDaytime(time) {
    if (isNaN(time[0])) {
        // assume daytime (relative time of day like sunrise, sunset)
        if (daytimeEvents[time] !== undefined) {
            // allow the keyword without delay
            return [time, true, '00:00:00'];
        }
        var delaySign = true,
            signIdx = -1;
        if ((signIdx = time.indexOf('-')) > 0) {
            delaySign = false; // negative delay
        } else if ((signIdx = time.indexOf('+')) <= 0) {
            LOG.logln('Error: invalid daytime format');
            return false;
        }
        var daytimeEvent = daytimeEvents[time.substr(0, signIdx)];
        if (daytimeEvent === undefined) {
            LOG.logln('Error: invalid daytime event `' + daytimeEvent + '\'');
            return false;
        }
        var delay = time.substr(signIdx + 1);
        return [daytimeEvent, delaySign, delay];
    }
    return time;
}

/**
 * Compute the sunrise and sunset time for a given date and the current system location
 * @param d {Date} The date to compute the times for
 * @return {Date} Date object or null if no system location is set or daytime is invalid.
 */
function getDaytimeForDate(daytime, d) {
    var ndGeo = Property.getNode('/config/geodata'),
        sun = new SunCalc(),
        res = new Date(d),
        lat, lon, off,
        day,mon,year,
        calc,
        t;

    if (daytimeEvents[daytime] === undefined) {
        LOG.logln('Error: invalid daytime `' + daytime + '\'');
        return null;
    }

    if (ndGeo) {
        try {
            lat = parseFloat(ndGeo.getChild('latitude').getValue(), 10);
            lon = parseFloat(ndGeo.getChild('longitude').getValue(), 10);
            // off = ndGeo.getChild('offsetGMT').getValue(); // wrong, offsetGMT is not updated
            // getTzOff is returned in minutes of difference, thus the -60
            off = d.getTimezoneOffset() / -60;
        } catch(e) {
            LOG.logln('Warning: device location is not set');
            return null;
        }
    } else {
        LOG.logln('Warning: device location is not set');
        return null;
    }

    // object with sunrise and sunset fields and hh:mm:ss values
    day = d.getDate();
    mon = d.getMonth() + 1;
    year = d.getFullYear();
    // THE NEGATIVE LONGITUDE IS INTENDED AND REQUIRED
    calc = sun.compute_current_date(lat, -lon, year, mon, day, off);
    LOG.logln('DEBUG: ' + year +'-'+ pad(mon) +'-'+ pad(day) +' sunrise: '+ calc.sunrise + ' sunset: ' + calc.sunset);
    if (calc[daytime].indexOf('[') >= 0) {
        // returns [Above] or [Below]
        LOG.logln('Warning: invalid sunrise/sunset calculation');
        return null;
    }

    // set result time
    t = calc[daytime].split(':');
    res.setHours(parseInt(t[0], 10));
    res.setMinutes(parseInt(t[1], 10));
    res.setSeconds(parseInt(t[2], 10));
    return res;
}

/**
 * Checks if the system version matches at least a given version
 * @param version Version to check against (in the format X.Y.Z)
 * @return true if the system version is at least the passed version,
     false otherwise, including the case where the version node doesn't exist
 */
function requireSystemVersion(version) {
    var ndSysVersion = Property.getNode('/system/version/version');
    if (ndSysVersion) {
        var sysVersion = ndSysVersion.getValue().split('.');
        var reqVersion = version.split('.');
        var len = sysVersion.length < reqVersion.length ? sysVersion.length : reqVersion.length;
        var i;
        for (i = 0; i < len; ++i) {
            var sys = parseInt(sysVersion[i], 10);
            var req = parseInt(reqVersion[i], 10);
            if (req < sys || (i === len - 1 && req === sys)) {
                return true;
            }
            if (req > sys) {
                return false;
            }
        }
    }
    return false;
}

function startup() {
    // Prepare app
    Property.load();
    LOG.logln('timed-events running');

    var ndVersion = Property.getNode('version');
    if (ndVersion) {
        var version = ndVersion.getValue().split('.');
        if (version.length > 1 && version[0] == '0' && parseInt(version[1], 10) < 9) {
            convertOldEvents();
        }
    }
    // requeue events;
    var ndEvents = Property.getNode('events');
    if (ndEvents) {
        var events = ndEvents.getChildren();
        var i;
        for (i = 0; i < events.length; ++i) {
            requeueEvent(events[i]);
        }

    } else {
        createNode('events');
    }

    //set version number
    Property.setProperty('version', '0.9.8');
    Property.setFlag('version', 'ARCHIVE', true);
    Property.store();
}

/**
 * A function that converts stored timed events from the old to the new version
 */
function convertOldEvents() {
    LOG.logln('Converting events from old timed events app');
    Property.load();
    var ndEvents = Property.getNode('events');
    var events = ndEvents.getChildren();
    var i;
    for (i = 0; i < events.length; ++i) {
        var ndOldParams = events[i].getChild('params'),
            oldParams,
            oldRecurrence,
            oldActions;

        if (!ndOldParams) {
            LOG.logln('Warning: no old params found to convert on old event ' + events[i].name);
            continue;
        }

        try {
            // LOG.logln('Debug: old params: ' + ndOldParams.getValue());
            oldParams = JSON.parse(ndOldParams.getValue());
            // LOG.logln('Debug: old recurrence: ' + decodeURIComponent(oldParams.recurrence));
            oldRecurrence = JSON.parse(decodeURIComponent(oldParams.recurrence));
            // LOG.logln('Debug: old actions: ' + decodeURIComponent(oldParams.actions));
            oldActions = JSON.parse(decodeURIComponent(oldParams.actions));
        } catch (e) {
            LOG.logln('Deleting old event '+ events[i].name +': failed to parse');
            deleteOldEvent(events[i]);
            continue;
        }

        //
        // validity checks
        //

        if (!oldRecurrence
            || !oldActions
            || !oldActions.actions
            || !(oldActions.actions.length > 0)
        ) {
            // delete invalid events
            LOG.logln('Deleting old event '+ events[i].name +': missing recurrence or actions');
            deleteOldEvent(events[i]);
            continue;
        }
        oldActions = oldActions.actions;

        if (! oldRecurrence.triggerTime
            || (oldRecurrence.recurrence == 'irregular'
                && (!oldRecurrence.dates
                    || oldRecurrence.dates.length == 0))
            || (oldRecurrence.recurrence == 'regular_wk'
                && (!oldRecurrence.weekdays
                    || oldRecurrence.weekdays.length == 0))
        ) {
            LOG.logln('Deleting old event '+ events[i].name +': missing time, dates or week days');
            deleteOldEvent(events[i]);
            continue;
        }

        //
        // Conversions
        //

        // new params object
        var params = {
            name: oldParams.name
        };

        // disabled event
        if (oldParams.Disabled === true) {
            params.disabled = true;
        }

        // Convert time from UTC to local time
        // (assuming correctly set timezone on the dSS)
        var utc = new Date();
        var time = oldRecurrence.triggerTime.split(':');
        utc.setUTCHours(parseInt(time[0], 10));
        utc.setUTCMinutes(parseInt(time[1], 10));
        utc.setUTCSeconds(parseInt(time[2], 10));
        params.time = // utc.format('H:i:s');
                    pad(utc.getHours())
                    + ':' + pad(utc.getMinutes())
                    + ':' + pad(utc.getSeconds());

        // recurrence period
        if (oldRecurrence.recurrence == 'regular_wk') {
            params.recurring = true;
        }

        if (params.recurring) {
            // weekdays
            params.recurrence = [];
            var j;
            for (j = 0; j < oldRecurrence.weekdays.length; ++j) {
                var idx = getDayIdx(oldRecurrence.weekdays[j]);
                if (idx !== -1) {
                    params.recurrence.push(idx);
                } else {
                    LOG.logln('Skipping invalid weekday: ' + oldRecurrence.weekdays[j]);
                    continue;
                }
            }
            if (params.recurrence.length == 0) {
                LOG.logln('Deleting old event '+ events[i].name +': no valid weekdays');
                deleteOldEvent(events[i]);
                continue;
            }

        } else {
            // reverse the dates from 13-01-2001 to 2001-01-13
            var dates = oldRecurrence.dates;
            var d;
            params.recurrence = [];
            for (d = 0; d < dates.length; ++d) {
                var date = dates[d].split('-');
                if (date.length != 3
                    || parseInt(date[1], 10) > 12) {
                    LOG.logln('Skipping invalid date');
                    continue;
                }
                params.recurrence.push(
                    pad(parseInt(date[2], 10)) + '-'
                    + pad(parseInt(date[1], 10)) + '-'
                    + pad(parseInt(date[0], 10))
                );
            }
            if (params.recurrence.length == 0) {
                LOG.logln('Deleting old event '+ events[i].name +': no valid dates');
                deleteOldEvent(events[i]);
                continue;
            }
        }

        // Convert old actions
        params.actions = [];
        var a;
        for (a = 0; a < oldActions.length; ++a) {
            var p = {};
            if (oldActions[a].Zone !== undefined) {
                p.zone = parseInt(oldActions[a].Zone, 10);
            }
            if (oldActions[a].Group !== undefined) {
                p.group = oldActions[a].Group === '' ?
                    0 : parseInt(oldActions[a].Group, 10);
            }
            if (oldActions[a].Scene !== undefined) {
                p.scene = parseInt(oldActions[a].Scene, 10);
            }
            if (oldActions[a].Output !== undefined) {
                p.outputValue = parseInt(oldActions[a].Output, 10);
            }
            if (oldActions[a].Device) {
                p.device = oldActions[a].Device;
            }
            if (oldActions[a].Blink) {
                p.blink = true;
            }
            params.actions.push(p);
        }

        // store converted event
        deleteOldEvent(events[i]);
        addEvent(params);
        LOG.logln('Converted event ' + params.idname);
    }
    Property.store();
} // convertOldEvents

/**
 * Execute actions delayed one by one
 * @param actions {Array} Actions to execute (in the order of the array)
 * @param time {Number} Miliseconds to wait between actions
 * @param callback {Function} (optional) callback to execute after the last action execution
 */
function executeActions(actions, time, callback) {
    var action = actions.shift();

    var target = null;
    if (action.group != null) {
        LOG.logln('target zone ' + action.zone + ' group ' + action.group);
        target = getDevices().byZone(action.zone).byGroup(action.group);

    } else if (action.device != null) {
        LOG.logln('target device ' + action.device);
        target = getDevices().byDSID(action.device);
    }

    if (target) {
        if (action.scene !== undefined) {
            LOG.logln('Exec: calling scene ' + action.scene);
            target.callScene(parseInt(action.scene, 10));
        }
        else if (action.blink) {
            LOG.logln('Exec: blinking');
            target.blink();
        }
        else if (action.outputValue !== undefined) {
            LOG.logln('Exec: setting ouput ' + action.outputValue);
            target.setValue(parseInt(action.outputValue, 10));
        }
    } else {
        LOG.logln('Warning: empty target');
    }

    if (actions.length > 0) {
        LOG.logln('Scheduling next '+ actions.length +' params in '+ time +'ms');
        var o = function () { executeActions(actions, time, callback); };
        setTimeout(time, o);
    } else {
        callback();
    }
} // executeActions

/**
 * Parses the recurrence object and generates an array of strings, each
 * containing a valid recurrence rule.
 * The dSS only accepts iCal dates if one provides both rrule and start time
 * @param params {Object} The event's parameters object
 * @return {Array} An array of rrule/starting time pair objects
 */
function generateIcalRules(params) {
    var rules = [], // return array
        daytime, // daytime array (only used if daytime event)
        delay = [0,0,0], // time delay only used on recurring daytime event (unsigned)
        delayMs = 0, // time delay in millisecs (signed)
        time = params.time.split(':'), // time array (delay array daytime event)
        timestring = '', // timestring used in the iCal rule
        d = new Date(), // the date object with scheduled date/time
        // use compatibility mode with old dSS versions (< 1.0.5)
        use_utc_time = !requireSystemVersion('1.0.5'),

        // loop variables:
        i,
        rrule = null,
        iCalStartTime = null;

    if (params.isDaytimeEvent) {
        daytime = getDaytime(params.time);
        time = daytime[2].split(':');
        delay = [   parseInt(time[0], 10),
                    parseInt(time[1], 10),
                    parseInt(time[2], 10)   ];
        delayMs = ((((delay[0] * 60) + delay[1]) * 60) + delay[2]) * 1000;
        if (!daytime[1]) {
            delayMs = -delayMs;
        }
    } else {
        d.setHours(parseInt(time[0], 10));
        d.setMinutes(parseInt(time[1], 10));
        d.setSeconds(parseInt(time[2], 10));
    }

    if (daytime) {
        // needs to be requeued after every execution
        rrule = 'FREQ=DAILY;COUNT=1';

        if (params.recurring) {
            // get the next date
            var day = d.getDay(),
                next = 7,
                now = new Date();

            // 60s ahead because it's too late to schedule an earlier event
            // if it's missed, the next event won't be scheduled until the
            // dSS is restarted and the queue rebuilt
            now.setTime(now.getTime() + (60 * 1000)); // add 60s in millisecs

            // loop to find the nearest scheduled weekday
            for (i = 0; i < params.recurrence.length; ++i) {
                // add 1 because our week starts on MO, JS Date's already on SU.
                // Also add 7 because JS modulo's can be < 0
                var n = (7 + params.recurrence[i] + 1 - day) % 7;
                if (n === 0) {
                    // nearest weekday is today: check if it's still possible to schedule it
                    var dtmp = getDaytimeForDate(daytime[0], d);
                    dtmp.setTime(dtmp.getTime() + delayMs);
                    LOG.logln('DEBUG: scheduling possible today: '
                        + (dtmp < now ? 'no' : 'yes')
                        + ' - ' + daytime[0]
                        + ' time + delay (' + dtmp.toString() + ') > now + 60s');
                    if (dtmp < now) {
                        // too late, next week then, but look for closer day
                        n = 7;
                    } else {
                        // right on time
                        next = 0;
                        break;
                    }
                }
                if (n < next) {
                    // found a closer day
                    next = n;
                }
            }
            // add day difference in millisecs
            d.setTime(d.getTime() + (next * 24 * 60 * 60 * 1000));

            d = getDaytimeForDate(daytime[0], d);
            d.setTime(d.getTime() + delayMs);

            iCalStartTime =
                d.getFullYear()
                + pad(d.getMonth() + 1) // goes from 0..11
                + pad(d.getDate())
                + 'T'
                + pad(d.getHours())
                + pad(d.getMinutes())
                + pad(d.getSeconds());

            rules.push({
                iCalRRule: rrule,
                iCalStartTTime: iCalStartTime
            });

        } else {
            // !recurring -> schedule every event independently
            for (i = 0; i < params.recurrence.length; ++i) {
                var dt = params.recurrence[i].split('-');
                d = new Date(
                    parseInt(dt[0], 10),
                    parseInt(dt[1], 10) - 1,
                    parseInt(dt[2], 10)
                );
                d = getDaytimeForDate(daytime[0], d);
                d.setTime(d.getTime() + delayMs);
                timestring =
                    pad(d.getHours())
                    + pad(d.getMinutes())
                    + pad(d.getSeconds());

               iCalStartTime = dt.join('') + 'T' + timestring;
               rules.push({
                   iCalRRule: rrule,
                   iCalStartTTime: iCalStartTime
               });
            }
        }
    } else {
        // !daytimed
        if (use_utc_time) {
            // run in compatibility mode
            timestring =
                pad(d.getUTCHours())
                + pad(d.getUTCMinutes())
                + pad(d.getUTCSeconds())
                + 'Z';
        } else {
            // use local time
            timestring =
                pad(d.getHours())
                + pad(d.getMinutes())
                + pad(d.getSeconds());
            // LOG.logln('Debug: event scheduled in local time for ' + d);
        }

        if (params.recurring) {

            iCalStartTime =
                d.getFullYear()
                + pad(d.getMonth() + 1) // goes from 0..11
                + pad(d.getDate())
                + 'T'
                + timestring;

            var weekdays = [];
            for (i = 0; i < params.recurrence.length; ++i) {
                weekdays.push(WEEK[params.recurrence[i]]);
            }
            rrule = 'FREQ=WEEKLY;BYDAY=' + weekdays;

            rules.push({
                iCalRRule: rrule,
                iCalStartTTime: iCalStartTime
            });

        } else {
            rrule = 'FREQ=DAILY;COUNT=1';
            for (i = 0; i < params.recurrence.length; ++i) {
                iCalStartTime = params.recurrence[i].replace(/-/g, '') + 'T' + timestring;
                rules.push({
                    iCalRRule: rrule,
                    iCalStartTTime: iCalStartTime
                });
            }
        }
    }
    // LOG.logln('Debug: rrule: `'+ rrule +'\', start time: ' + iCalStartTime);

    return rules;
}

/**
 * Remove all queued occurences of this event
 * @param ndEvent The event node of the event to reqeue
 */
function dequeueEvent(ndEvent) {
    var ndQueue = ndEvent.getChild('queueIDs');
    if (!ndQueue) {
        return;
    }
    var queue = ndQueue.getChildren();
    var i;
    for (i = 0; i < queue.length; ++i) {
        var propNode = Property.getNode(
            '/system/EventInterpreter/ScheduledEvents/' + queue[i].getValue()
        );
        if (propNode) {
            propNode.getParent().removeChild(propNode);
        }
        ndQueue.removeChild(queue[i]);
    }
}

/**
 * Queue an event
 * @param ndEvent The event node of the event to reqeue
 * @param params (optional) pre-loaded params of the node
 */
function queueEvent(ndEvent, params) {
    var qIdPath = 'events/' + ndEvent.name + '/queueIDs';
    var ndQueue = ndEvent.getChild('queueIDs') || createNode(qIdPath); // just here to ensure the node is created
    var i;

    // raise the event and store the queue id
    if (!params) {
        params = loadParams(ndEvent);
    }
    if (params.disabled) {
        return;
    }
    LOG.logln('Debug: queueing events for event ' + params.idname);

    var ical = generateIcalRules(params);
    for (i = 0; i < ical.length; ++i) {
        var timerEvent = new TimedICalEvent(
            'timed-events',
            ical[i].iCalStartTTime,
            ical[i].iCalRRule,
            {
                event_id: params.id,
                action_type: 'execute'
            }
        );
        LOG.logln('Debug: - ' + ical[i].iCalStartTTime);
        Property.setProperty(qIdPath +'/'+ i, timerEvent.raise());
    }
}

/**
 * Check if all occurences of an event are in the queue, else requeue them
 * @param ndEvent The event node of the event to reqeue
 * @param params (optional) pre-loaded params of the node
 */
function requeueEvent(ndEvent, params) {
    var ndQueue = ndEvent.getChild('queueIDs');
    var i, queue, propNode;

    if (!ndQueue) {
        queueEvent(ndEvent, params);

    } else {
        queue = ndQueue.getChildren();
        if (queue.length == 0) {
            queueEvent(ndEvent, params);

        } else {
            for (i = 0; i < queue.length; ++i) {
                propNode = Property.getNode(
                    '/system/EventInterpreter/ScheduledEvents/' + queue[i].getValue()
                );
                if (!propNode) {
                    // if any node doesn't exists, delete and requeue them all
                    dequeueEvent(ndEvent);
                    queueEvent(ndEvent, params);
                    break;
                }
            }
        }
    }
}

/**
 * Fetches the parameters of this event from the property tree.
 * @param eventNode {Object} Property tree node of the event
 * @return {Object} Object containing the read out values of the event
 */
function loadParams(eventNode) {
    LOG.logln('Loading params for node with id: ' + eventNode.name);
    var name = eventNode.getChild('name');
    if (!name) {
        LOG.logln('Error: no event name');
        return;
    }

    var params = {
        id: eventNode.name,
        idname: eventNode.name + ' (' + name.getValue() + ')',
        name: name.getValue(),
        time: eventNode.getChild('time').getValue(),
        actions: [],
        recurrence: []
    };
    var ndDisabled = eventNode.getChild('disabled');
    params.disabled = (ndDisabled ? !! ndDisabled.getValue() : false);

    // set daytime flag if the time is sunrise+hh:mm:ss or sunset-hh:mm:ss
    // check true if the first char of time is not a number
    params.daytime = (isNaN(params.time[0]));

    // LOG.logln('Debug: Loading actions');
    var ndActions = eventNode.getChild('actions');
    var loadValue = function(action, ndAction, name) {
        var c = ndAction.getChild(name);
        if (c) {
            action[name] = c.getValue();
            return true;
        }
        return false;
    };
    if (ndActions) {
        var actions = ndActions.getChildren();
        var a;
        for (a = 0; a < actions.length; ++a) {
            var curAction = {};
            if (loadValue(curAction, actions[a], 'device')) {}
            else if (loadValue(curAction, actions[a], 'group') && loadValue(curAction, actions[a], 'zone')) {}
            else {
                LOG.logln('Error: event '+ params.idname +' has an action with no target');
                continue;
            }
            if (loadValue(curAction, actions[a], 'blink')) {}
            else if (loadValue(curAction, actions[a], 'scene')) {}
            else if (loadValue(curAction, actions[a], 'outputValue')) {}
            else {
                LOG.logln('Error: event '+ params.idname +' has an action with invalid action to take');
                continue;
            }
            params.actions.push(curAction);
            // LOG.logln('Debug: action: '+ JSON.stringify(curAction));
        }
    } else {
        LOG.logln('Error: event '+ params.idname +' has no actions set');
    }

    // LOG.logln('Debug: loading recurrences');
    var ndRecurring = eventNode.getChild('recurring');
    params.recurring = (ndRecurring ? !!ndRecurring.getValue() : false);

    var ndRecurrence = eventNode.getChild('recurrence');
    if (ndRecurrence) {
        var recurrences = ndRecurrence.getChildren();
        if (params.recurring) {
            // weekdays
            var w;
            for (w = 0; w < recurrences.length; ++w) {
                params.recurrence.push(recurrences[w].name);
            }
        } else {
            // dates
            var d;
            for (d = 0; d < recurrences.length; ++d) {
                params.recurrence.push(recurrences[d].getValue());
            }
        }
        // LOG.logln('Debug: recurrences: '+ JSON.stringify(params.recurrence));
    } else {
        LOG.logln('Error: event '+ params.idname +' has no recurrence set');
    }
    // LOG.logln('Debug: params loaded');
    return params;
}

/**
 * Updates a node if it exists or creates a new archived property node otherwise
 * @param path {String} Relative path
 * @param value {String|Boolean|Integer} Value to set
 */
function updateOrCreateNode(path, value) {
    var nd = Property.getNode(path);
    if (nd) {
        nd.setValue(value);
    } else {
        createNode(path, value);
    }
}

/**
 * Creates a new archived property node with given value
 * @param path {String} Relative path
 * @param value {String|Boolean|Integer} Value to set
 * @return {Object} The newly created node
 */
function createNode(path, value) {
    var nd = new Property(path);
    if (value !== undefined) {
        nd.setValue(value);
    }
    nd.setFlag('ARCHIVE', true);
    return nd;
}

/**
 * Stores _VALID_ parameters into the property tree.
 * Use checkParams(params) to validate them first!
 * Make sure to call Property.store() after calling this method
 * @param params {Object} Object to store
 */
function storeParams(params) {
    var eventPath = 'events/' + params.id;
    var actionPath = eventPath + '/actions';
    var i, c;

    var ndEvent = Property.getNode(eventPath) || createNode(eventPath);

    updateOrCreateNode(eventPath + '/id', params.id);
    updateOrCreateNode(eventPath + '/name', params.name);
    updateOrCreateNode(eventPath + '/time', params.time);
    updateOrCreateNode(eventPath + '/disabled', !!params.disabled);

    // actions
    var ndActions = ndEvent.getChild('actions');
    if (!ndActions) {
        ndActions = createNode(actionPath);
    } else {
        // clean out actions, as indices might not match up
        for (c in ndActions.getChildren()) {
            ndActions.removeChild(c);
        }
    }
    // populate actions
    for (i = 0; i < params.actions.length; ++i) {
        var p = actionPath + '/' + i;
        createNode(p);

        if (params.actions[i].zone !== undefined) {
            createNode(p + '/zone', params.actions[i].zone);
        }
        if (params.actions[i].group !== undefined) {
            createNode(p + '/group', params.actions[i].group);
        }
        if (params.actions[i].device) {
            createNode(p + '/device', params.actions[i].device);
        }
        if (params.actions[i].blink) {
            createNode(p + '/blink', true);
        }
        if (params.actions[i].scene !== undefined) {
            createNode(p + '/scene', params.actions[i].scene);
        }
        if (params.actions[i].outputValue !== undefined) {
            createNode(p + '/outputValue', params.actions[i].outputValue);
        }
    }

    // recurrence
    updateOrCreateNode(eventPath + '/recurring', params.recurring);

    var ndRecurrence = ndEvent.getChild('recurrence')
                    || createNode(eventPath + '/recurrence');
    if (params.recurring) {
        // weed out existing weekdays
        for (c in ndRecurrence.getChildren()) {
            ndRecurrence.removeChild(c);
        }
        for (i = 0; i < params.recurrence.length; ++i) {
            var idx = params.recurrence[i];
            createNode(eventPath + '/recurrence/' + idx, WEEK[idx]);
        }

    } else {
        // weed out dates
        for (c in ndRecurrence.getChildren()) {
            ndRecurrence.removeChild(c);
        }
        for (i = 0; i < params.recurrence.length; ++i) {
            createNode(eventPath + '/recurrence/' + i, params.recurrence[i]);
        }
    }
    return ndEvent;
} // storeParams

/**
 * Checks if parameters conform to what is expected
 * @param eventParams {Object} Object to check
 * @return {Object} Conform object or null on error
 */
function checkParams(eventParams) {
    if (!eventParams.name) {
        LOG.logln('Error: event has no name');
        return null;
    }

    if (!eventParams.actions || eventParams.actions.length === 0) {
        LOG.logln('Error: event has no actions');
        return null;
    }

    if (!eventParams.recurrence || eventParams.recurrence.length === 0) {
        LOG.logln('Error: event has no recurrence');
        return null;
    }

    if (!eventParams.time) {
        LOG.logln('Error: event has no time set');
        return null;
    }

    var actions = [];
    var i;
    for (i = 0; i < eventParams.actions.length; ++i) {
        var a = eventParams.actions[i];
        if (!a.device && isNaN(a.group)) { // isNan(undefined) === true
            LOG.logln('Error: no or invalid group or device defined as target on action ' + i);
            return null;
        }
        if (a.blink === undefined && a.scene === undefined && a.outputValue === undefined) {
            LOG.logln('Error: no action defined at index ' + i);
            return null;
        }
        if (
            (a.scene !== undefined && isNaN(a.scene))
            || (a.group !== undefined && isNaN(a.group))
            || (a.outputValue !== undefined && (
                isNaN(a.outputValue) || a.outputValue < 0 || a.outputValue > 255
            ))
        ) {
            LOG.logln('Error: invalid action defined at index ' + i);
            return null;
        }

        var action = {};
        if (a.zone !== undefined) {
            action.zone = parseInt(a.zone, 10);
        }
        if (a.group !== undefined) {
            action.group = parseInt(a.group, 10);
        }
        if (a.device !== undefined) {
            action.device = a.device;
        }
        if (a.blink) {
            action.blink = true;
        }
        if (a.scene !== undefined) {
            action.scene = parseInt(a.scene, 10);
        }
        if (a.outputValue !== undefined) {
            action.outputValue = parseInt(a.outputValue, 10);
        }
        actions.push(action);
    }

    var recurring = !!eventParams.recurring;
    var recurrence = [];
    var d;
    if (recurring) {
        for (i = 0; i < eventParams.recurrence.length; ++i) {
            d = eventParams.recurrence[i];
            if (isNaN(d) || d < 0 || d > 6) {
                LOG.logln('Warning: wrong weekday index `'+ d +'\', skipping');
                continue;
            }
            recurrence.push(d);
        }
    } else {
        for (i = 0; i < eventParams.recurrence.length; ++i) {
            d = eventParams.recurrence[i];
            var date = d.split('-');
            if (date.length !== 3 || isNaN(date[0]) || isNaN(date[1]) || isNaN(date[2])
                || date[0].length < 4 || date[1] == 0 || date[1] > 12
                || date[2] == 0 || date[2] > 31
            ) {
                LOG.logln('Warning: invalid date `'+ d +'\'');
                continue;
            }
            recurrence.push(d);
        }
    }
    if (recurrence.length == 0) {
        LOG.logln('Error: no valid recurrence');
        return null;
    }
    var time = eventParams.time;
    var daytime = getDaytime(eventParams.time);
    var daytimeEvent = '';
    var isDaytimeEvent = false;
    if (daytime === false) {
        return null; // error was logged in getDaytime
    } else if (typeof(daytime) !== 'string') {
        isDaytimeEvent = true;
        daytimeEvent = daytime[0] + (daytime[1] ? '+' : '-');
        time = daytime[2];
        if (!requireSystemVersion('1.2.0')) {
            LOG.logln('Warning: this dSS version does not support daytimed events, this event will be disabled - please update your dSS');
            eventParams.disabled = true;
        }
    }
    var t = time.split(':');
    var inv = (t.length !== 3);
    for (i = 0; !inv && i < 3; ++i) {
        if (!isNaN(t[i])) {
            t[i] = parseInt(t[i], 10);
            inv = inv || t[i] < 0 || t[i] > 59;
        } else {
            inv = true;
        }
    }
    inv = inv || (!isDaytimeEvent && t[0] > 23);
    if (inv) {
        LOG.logln('Error: invalid time format');
        return null;
    }
    time = daytimeEvent + pad(t[0]) + ':' + pad(t[1]) + ':' + pad(t[2]);

    // create the valid params object with all fields except id and idname
    var params = {
        name: eventParams.name,
        actions: actions,
        disabled: !!eventParams.disabled,
        recurring: recurring,
        recurrence: recurrence,
        isDaytimeEvent: isDaytimeEvent,
        time: time
    };
    // add id, idname fields if they were supplied
    if (eventParams.id) {
        params.id = eventParams.id;
        params.idname =  params.id + ' (' + params.name + ')';
    }

    return params;
} // checkParams

/**
 * Add a new event
 * @param params {Object} The new event is created from these parameters
 */
function addEvent(eventParams) {
    // check eventParams and set params for serialization
    var params = checkParams(eventParams);
    if (!params) {
        return;
    }

    if (params.id === undefined) {
        params.id = getId();
        params.idname = params.id + ' (' + params.name + ')';
    }

    var ndEvent = storeParams(params);

    if (!params.disabled) {
        queueEvent(ndEvent, params);
    }

    // set archive flags
    Property.setFlag('events/' + params.id, 'ARCHIVE', true);
    Property.store();
}

function executeEvent(eventNode, params) {
    if (!params.disabled) {
        LOG.logln(
            'Processing event '+ params.idname
            + ' (' + params.actions.length +' actions)'
        );
        executeActions(params.actions, 500, function() {
            disableOnLastOccurence(eventNode, params);
        });
    }
}

function disableOnLastOccurence(eventNode, params) {
    // Disable irregular event if it's its last occurrence
    if (!params.recurring) {
        if (eventNode.getChild('queueIDs').getChildren().length === 1) {
            enableEvent(eventNode, params, false);

            // notify UI about changes
            var updateEvent = new Event('timed-events.notifyUi', params.id);
            updateEvent.raise();
        }
    }
}

function deleteOldEvent(eventNode) {
    eventNode.getParent().removeChild(eventNode);
    Property.store();
}

function deleteEvent(eventNode, params) {
    LOG.logln('Delete: event '+ params.idname);

    dequeueEvent(eventNode);
    eventNode.getParent().removeChild(eventNode);
    Property.store();
}

/**
 * Enable or disable a timed event
 * @param enable {Boolean} True to enable, false to disable
 */
function enableEvent(eventNode, params, enable) {
    LOG.logln((enable ? 'Enabling' : 'Disabling') + ' event ' + params.idname);

    params.disabled = !enable;
    updateOrCreateNode('events/' + params.id + '/disabled', !enable);
    if (enable) {
        requeueEvent(eventNode, params);
    } else {
        dequeueEvent(eventNode);
    }
    Property.store();
}

function main() {
    if (raisedEvent.name == 'running') {
        startup();
        return;
    }
    /*
    LOG.logln('Debug: timed-events main call with action type: '
                + raisedEvent.parameter.action_type
                + (raisedEvent.parameter.params === undefined ? ''
                    : ' and params: '
                    + raisedEvent.parameter.params.toString()
                )
    );
    */
    var params;
    if (raisedEvent.parameter.action_type == 'add') {
        // Add a new event
        if (raisedEvent.parameter.params === undefined) {
            LOG.logln('Error: missing event parameters');
            return;
        }
        LOG.logln('Adding event with parmeters: '+ raisedEvent.parameter.params);
        try {
            params = JSON.parse(raisedEvent.parameter.params);
            addEvent(params);
        } catch(e) {
            LOG.logln('Error: Failed to parse params');
        }

    } else {
        // Operations that involve a specific event's parameters
        Property.load();
        var ndEvent = Property.getNode('events/' + raisedEvent.parameter.event_id);
        if (ndEvent) {
            params = loadParams(ndEvent);

            if (raisedEvent.parameter.action_type == 'enable') {
                enableEvent(ndEvent, params, true);
            }
            else if (raisedEvent.parameter.action_type == 'disable') {
                enableEvent(ndEvent, params, false);
            }
            else if (raisedEvent.parameter.action_type == 'delete') {
                deleteEvent(ndEvent, params);
            }
            else if (raisedEvent.parameter.action_type == 'execute') {
                executeEvent(ndEvent, params);
            }
        } else {
            LOG.logln('Error: node with id '+ raisedEvent.parameter.event_id +' doesn\'t exist');
        }
    }
} // main

main();

